# MIT License

# Copyright (c) 2021 Northern.tech

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import sys
import json

from shutil import rmtree
from collections import OrderedDict
from typing import Union

from glrp.pretty import pretty


def find(name, recursive=True, directories=False, files=True, extension=None):
    assert files or directories
    assert os.path.isdir(name)
    for root, subdirs, subfiles in os.walk(name):
        if directories:
            for dir in subdirs:
                if not extension or (extension and dir.endswith(extension)):
                    yield os.path.join(root, dir) + "/"
        if files:
            for file in subfiles:
                if not extension or (extension and file.endswith(extension)):
                    yield os.path.join(root, file)
        if not recursive:
            return  # End iteration after looking through first (top) level


def rm(path: str, missing_ok=False):
    if not missing_ok:
        assert os.path.exists(path)
    if missing_ok and not os.path.exists(path):
        return False
    if os.path.isdir(path):
        rmtree(path)
    else:  # Assume path is a file
        os.remove(path)  # Will raise exception if missing
    return True


def mkdir(path: str, exist_ok=True):
    os.makedirs(path, exist_ok=exist_ok)


def read_file(path):
    try:
        with open(path, "r") as f:
            return f.read()
    except FileNotFoundError:
        return None


def save_file(path, data):
    if "/" in path:
        mkdir("/".join(path.split("/")[0:-1]))
    with open(path, "w") as f:
        f.write(data)


def read_json(path) -> Union[OrderedDict, None]:
    try:
        with open(path, "r") as f:
            return json.loads(f.read(), object_pairs_hook=OrderedDict)
    except FileNotFoundError:
        return None
    except NotADirectoryError:
        return None
    except json.decoder.JSONDecodeError as ex:
        print("Error reading json file '{}': {}".format(path, ex))
        sys.exit(1)


def write_json(path, data):
    data = pretty(data) + "\n"
    return save_file(path, data)



import json
from glrp.summary import CommitSummary, Commit, Person


def test_summary_to_dict_empty():
    summary = CommitSummary()

    expected = {
        "counts": {
            "commits": 0,
            "signed": 0,
            "unsigned": 0,
            "trusted": 0,
            "untrusted": 0,
        },
        "emails": {},
        "names": {},
        "fingerprints": {},
        "ids": {},
    }

    assert summary.to_dict() == expected


def test_person():
    person = Person("John Doe <john.doe@example.com>")
    assert person.email == "john.doe@example.com"
    assert person.name == "John Doe"
    assert person.id == "John Doe <john.doe@example.com>"


def test_commit():
    author = Person("John Doe <john.doe@example.com>")
    committer = Person("John Doe <john.doe@example.com>")
    commit = Commit(author, committer, "Initial commit")
    assert commit.author.id == "John Doe <john.doe@example.com>"
    assert commit.committer.id == "John Doe <john.doe@example.com>"
    assert commit.message == "Initial commit"
    assert commit.fingerprint == None

    author = Person("Alice <alice@example.com>")
    committer = Person("Bob <bob@example.com>")
    commit = Commit(
        author, committer, "Second commit", "AFE8C5F43057C0093122299F584211AF6AB3EE12"
    )
    assert commit.author.id == "Alice <alice@example.com>"
    assert commit.author.name == "Alice"
    assert commit.author.email == "alice@example.com"
    assert commit.committer.id == "Bob <bob@example.com>"
    assert commit.committer.name == "Bob"
    assert commit.committer.email == "bob@example.com"
    assert commit.message == "Second commit"
    assert commit.fingerprint == "AFE8C5F43057C0093122299F584211AF6AB3EE12"


def test_commit_summary():
    john = Person("John Doe <john.doe@example.com>")
    commit = Commit(john, john, "Initial commit")
    summary = CommitSummary(commit)
    assert summary.to_dict() == {
        "counts": {
            "commits": 1,
            "signed": 0,
            "unsigned": 1,
            "trusted": 0,
            "untrusted": 0,
        },
        "emails": {
            "john.doe@example.com": {
                "counts": {
                    "commits": 1,
                    "signed": 0,
                    "unsigned": 1,
                    "trusted": 0,
                    "untrusted": 0,
                },
                "names": ["John Doe"],
                "ids": ["John Doe <john.doe@example.com>"],
                "fingerprints": [],
            }
        },
        "names": {
            "John Doe": {
                "counts": {
                    "commits": 1,
                    "signed": 0,
                    "unsigned": 1,
                    "trusted": 0,
                    "untrusted": 0,
                },
                "emails": ["john.doe@example.com"],
                "ids": ["John Doe <john.doe@example.com>"],
                "fingerprints": [],
            }
        },
        "fingerprints": {},
        "ids": {
            "John Doe <john.doe@example.com>": {
                "counts": {
                    "commits": 1,
                    "signed": 0,
                    "unsigned": 1,
                    "trusted": 0,
                    "untrusted": 0,
                },
                "names": ["John Doe"],
                "emails": ["john.doe@example.com"],
                "fingerprints": [],
            }
        },
    }


def test_commit_summary_json():
    john = Person("John Doe <john.doe@example.com>")
    commit = Commit(john, john, "Initial commit")
    summary = CommitSummary(commit)
    dump = str(summary)
    data = json.loads(dump)
    assert data == summary.to_dict()



import subprocess
import os
import sys
import argparse
import json
from typing import Optional, List

from glrp.internal_parser import parse, parse_to_all_representations
from glrp.version import string as version_string
from glrp.utils import find, mkdir, rm, write_json, read_json
from glrp.pretty import pretty as prettify
from glrp.summary import CommitSummary, Commit
from glrp.compare import compare_summaries

# Usage:
# git log -p --format=raw --show-signature --stat | glrp
#
# glrp .
#
# glrp --compare main~2,main...main~2
#
# glrp --combine .before.json,.after.json

all_processes = []


class GlobalState:
    def __init__(self):
        self.summary = CommitSummary()  # Built up by adding
        self.commits = {}  # All commits, keyed by SHA
        self.trusted = None  # Set by --trusted argument

    def _get_fingerprints(self, folder: str):
        assert os.path.isdir(folder)
        for file in find(folder, extension=".fp"):
            with open(file, "r") as f:
                for line in f:
                    line = line.strip()
                    line = line.replace(" ", "")
                    if line:
                        yield line

    def set_trusted_fingerprints(self, folder: str):
        self.trusted = list(self._get_fingerprints(folder))
        print(f"Trusting {len(self.trusted)} fingerprints")
        for fingerprint in self.trusted:
            print(fingerprint)

    def record_commit(self, commit):
        self.commits[commit["commit"]] = commit

        commit_obj = Commit.from_json(commit)
        summary = CommitSummary(commit_obj, trusted=self.trusted)
        self.summary = self.summary + summary


global_state = GlobalState()


def output_to_directory(output_dir):
    assert output_dir is not None and output_dir != ""
    if not output_dir.endswith("/"):
        output_dir = output_dir + "/"
    if (
        not output_dir.startswith("./")
        and not output_dir.startswith("/")
        and not output_dir.startswith("~/")
    ):
        output_dir = "./" + output_dir

    assert output_dir != "/"
    assert output_dir != "./"
    assert output_dir != "~/"
    assert output_dir != "."

    assert os.path.isdir(output_dir) or not os.path.exists(output_dir)

    rm(output_dir, missing_ok=True)

    mkdir(f"{output_dir}", exist_ok=True)

    with open(f"{output_dir}summary.json", "w") as f:
        f.write(prettify(global_state.summary.to_dict()) + "\n")

    index = 0
    for sha, commit in global_state.commits.items():
        write_json(
            f"{output_dir}shas/{sha}.json",
            commit,
        )
        write_json(
            f"{output_dir}index/{str(index).rjust(6, '0')}.json",
            commit,
        )
        index += 1


def dump_commit(raw_commit, split_commit, pretty_commit):
    sha = pretty_commit["commit"]
    with open(f"./debug/{sha}.1.raw.txt", "w") as f:
        f.write("\n".join(raw_commit))
    with open(f"./debug/{sha}.2.raw.json", "w") as f:
        f.write(prettify(raw_commit))
    with open(f"./debug/{sha}.3.split.json", "w") as f:
        f.write(prettify(split_commit))
    with open(f"./debug/{sha}.4.pretty.json", "w") as f:
        f.write(prettify(pretty_commit))


def _validate(
    input: Optional[str] = None,
    output: Optional[str] = None,
    output_dir: Optional[str] = None,
    quiet: bool = False,
    debug: bool = False,
    summarize: bool = False,
    pretty: bool = False,
):
    assert (
        input is None or input == "-" or os.path.isfile(input) or os.path.isdir(input)
    )
    if output is not None:
        assert isinstance(output, str) and len(output) > 0
        assert os.path.isfile(output) or not os.path.exists(output)
    if output_dir is not None:
        assert isinstance(output_dir, str) and len(output_dir) > 0
        assert os.path.isdir(output_dir) or not os.path.exists(output_dir)
    assert quiet is True or quiet is False
    assert debug is True or debug is False
    assert summarize is True or summarize is False
    assert pretty is True or pretty is False
    return


def _parse_logs(
    input,
    output_dir: Optional[str],
    quiet: bool,
    debug: bool,
    summarize: bool,
    pretty: bool,
    trusted: Optional[str],
):
    if trusted:
        global_state.set_trusted_fingerprints(trusted)
    if debug:
        rm("./debug/", missing_ok=True)
        mkdir("./debug/", exist_ok=False)
        for raw_commit, split_commit, pretty_commit in parse_to_all_representations(
            input
        ):
            dump_commit(raw_commit, split_commit, pretty_commit)
            if not quiet:
                if pretty:
                    print(prettify(pretty_commit))
                else:
                    print(json.dumps(pretty_commit))
        return

    for commit in parse(input):
        if not summarize and not quiet:
            if pretty:
                print(prettify(commit))
            else:
                print(json.dumps(commit))
        if summarize or output_dir:
            global_state.record_commit(commit)

    if not summarize and not output_dir:
        return

    if summarize:
        assert not output_dir
        r = global_state.summary.to_dict()
        if not quiet:
            print(prettify(r))
        return r

    assert output_dir
    output_to_directory(output_dir)


class UserError(Exception):
    pass


def parse_logs(
    input: Optional[str] = None,
    output_dir: Optional[str] = None,
    quiet: bool = False,
    debug: bool = False,
    summarize: bool = False,
    pretty: bool = False,
    git_extra: Optional[List[str]] = None,
    trusted: Optional[str] = None,
):
    _validate(
        input=input,
        output_dir=output_dir,
        quiet=quiet,
        debug=debug,
        summarize=summarize,
        pretty=pretty,
    )

    if input in (None, "-"):
        input_file = sys.stdin
    else:
        if os.path.isdir(input):
            process = subprocess.Popen(
                (
                    ["git", "log", "-p", "--format=raw", "--show-signature", "--stat"]
                    + (git_extra if git_extra else [])
                ),
                cwd=input,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
            )
            all_processes.append(process)
            input_file = process.stdout
        else:
            try:
                input_file = open(input, "r")
            except:
                raise UserError(f"Could not open '{input}'.")

    r = _parse_logs(
        input=input_file,
        output_dir=output_dir,
        quiet=quiet,
        debug=debug,
        summarize=summarize,
        pretty=pretty,
        trusted=trusted,
    )
    for process in all_processes:
        process.wait()
    return r


def get_args():
    parser = argparse.ArgumentParser(
        prog="glrp",
        description="Parses the output of 'git log -p --format=raw --show-signature --stat'",
    )
    parser.add_argument(
        "input",
        type=str,
        nargs="?",
        help="File to read input from or folder to run 'git' in",
    )
    parser.add_argument("--version", action="version", version=version_string())
    parser.add_argument(
        "-o", "--output-dir", type=str, help="Output commits to a folder structure"
    )
    parser.add_argument(
        "--trusted",
        type=str,
        help="Path to folder of trusted GPG keys, stored in .fp files",
    )
    parser.add_argument(
        "-q",
        "--quiet",
        default=False,
        action="store_true",
        help="Stop printing JSON commits to standard out",
    )
    parser.add_argument(
        "-d",
        "--debug",
        action="store_true",
        help="Enable debug information",
    )
    parser.add_argument(
        "--summarize",
        default=False,
        action="store_true",
        help="Print a JSON summary instead of individual commits",
    )
    parser.add_argument(
        "--compare",
        type=str,
        help="Two comma separated commit hashes / ranges to compare",
    )
    parser.add_argument(
        "--combine",
        type=str,
        help="Comma separated list of summary filenames to combine",
    )
    parser.add_argument(
        "--pretty",
        default=False,
        action="store_true",
        help="Print commit JSONs on multiple lines, with indentation",
    )
    args = parser.parse_args()
    return args


def validate_args(args):
    if args.summarize and (args.compare or args.combine):
        raise UserError(
            "The --summarize option cannot be used with --compare or --combine"
        )
    if args.compare and args.combine:
        raise UserError("The --combine option cannot be used with --compare")
    if args.input and (args.combine or args.compare):
        raise UserError("The input argument cannot be used with --compare or --combine")
    if args.combine and (len(args.combine) < 3 or "," not in args.combine):
        raise UserError(
            "The --combine option requires two or more comma separated JSON filenames"
        )
    if args.trusted:
        if not os.path.exists(args.trusted):
            raise UserError(
                "The --trusted argument must be a path to a folder containing .fp files"
                + f" - '{args.trusted}' does not exist"
            )
        if not os.path.isdir(args.trusted):
            raise UserError(
                "The --trusted argument must be a path to a folder containing .fp files"
                + f" - '{args.trusted}' is not a directory"
            )
        files = find(
            args.trusted,
            recursive=False,
            directories=False,
            files=True,
            extension=".fp",
        )
        if next(files, None) is None:
            raise UserError(
                "The --trusted argument must be a path to a folder containing .fp files"
                + f" - No .fp files found in '{args.trusted}'"
            )
    return


def get_summary(ref: str | List[str]):
    global global_state
    if isinstance(ref, str) and ref.endswith(".json"):
        r = read_json(ref)
        return r
    if isinstance(ref, str):
        git_extra = [ref]
    else:
        git_extra = ref
    r = parse_logs(
        input=".",
        output_dir=None,
        quiet=True,
        debug=False,
        summarize=True,
        pretty=False,
        git_extra=git_extra,
    )
    global_state = GlobalState()
    return r


def _compare_commits(a, b):
    before = get_summary(a)
    assert before is not None
    write_json(".before.json", before)
    print(f"Saved .before.json with stats from {before['counts']['commits']} commits")
    after = get_summary(b)
    assert after is not None
    write_json(".after.json", after)
    print(f"Saved .after.json with stats from {after['counts']['commits']} commits")
    compare_summaries(before, after)


def intify(s: str) -> int | None:
    try:
        return int(s)
    except ValueError:
        return None


def compare_commits(compare):
    if "," in compare:
        a, b = compare.split(",")
        a = [a]
        b = [b]
        return _compare_commits(a, b)
    n = intify(compare)
    if n is not None:
        a = [f"--skip={n}"]
        b = [f"--max-count={n}"]
        return _compare_commits(a, b)
    a = [f"--until={compare}"]
    b = [f"--since={compare}"]
    return _compare_commits(a, b)


def combine_summaries(filenames):
    summaries = [CommitSummary(filename=f) for f in filenames.split(",")]
    combined = summaries[0]
    for summary in summaries[1:]:
        combined = combined + summary
    print(prettify(combined.to_dict()))


def main():
    args = get_args()
    validate_args(args)
    if args.compare:
        compare_commits(args.compare)
        return
    if args.combine:
        combine_summaries(args.combine)
        return
    parse_logs(
        input=args.input,
        output_dir=args.output_dir,
        quiet=args.quiet,
        debug=args.debug,
        summarize=args.summarize,
        pretty=args.pretty,
        trusted=args.trusted,
    )


if __name__ == "__main__":
    main()



"""
This file contains the internal / library functions for parsing the output
of 'git log -p --format=raw --show-signature --stat'
There is intentionally no main function / CLI here, only testable and importable
functions to perform the parsing. CLI, and additional tracking, summaries etc.
are implemented in git_log_raw_parser.py

This is done in 4 steps:
1. Iterate over the input stream generating lines
2. Iterate over lines, find where the next commit starts, and generate "raw commits" (list of lines within one commit)
3. Iterate over raw commits, find where each section starts, and generate "split commits"
4. Iterate over the split commits, generating pretty commits.

All of this is implemented using generators / iterators, so you can process and output results
while later commits are still being printed / processed.

Performance-wise, the 4 step parsing outlined above is not ideal, but it helps a lot
in making the code simpler to write and read and less error prone.
In pracitce, performance is good enough, even for hundreds, or thousands commits.

If you want to operate on the more raw data without much translations / prettifying,
you can skip step 4, or steps 3 and 4.

Note: The output of the git log command we use varies based on gpg (which keys are imported).
      Don't expect the same output on different machines with different gpg states.

Note: The parsing is lossy, we reconfigure the input stream to replace unicode
      errors (with question marks), so don't expect the diffs or commit messages to be
      100% correct for these cases.
"""

import sys


def _reconfigure(input_stream):
    # input_stream / sys.stdin is defined as TextIO in typeshed even though it's actually TextIOWrapper by default
    # This decision was made because it's common to reassign stdin / stdout / stderr:
    # https://github.com/python/typeshed/issues/10093
    # For this parser we need to use reconfigure from TextIOWrapper, so we don't
    # encounter exceptions for commit messages / diffs with weird characters.
    # So we need to assume that sys.stdin is actually TextIOWrapper,
    # the assertion makes this assumption explicit and makes strict typecheckers like Pyright happy.
    # assert isinstance(input_stream, TextIOWrapper)
    input_stream.reconfigure(errors="replace")


# Utilities:


def remove_prefix(line, prefix):
    assert line.startswith(prefix)
    offset = len(prefix)
    assert offset > 0
    return line[offset:]


def remove_suffix(line, suffix):
    assert line.endswith(suffix)
    offset = len(suffix)
    assert offset > 0
    return line[0:-offset]


def parse_author(line):
    author = {}
    split = line.split(" ")
    timezone = split[-1]
    timestamp = split[-2]
    suffix = " " + " ".join(split[-2:])
    full = remove_suffix(line, suffix)
    author["id"] = full
    assert full[-1] == ">"
    full = full[0:-1]
    split = full.split("<")
    assert len(split) == 2
    author["name"] = split[0].strip()
    author["email"] = split[1].strip()
    author["timestamp"] = timestamp
    author["timezone"] = timezone
    return author


def valid_signature(commit):
    for line in commit["gpg"]:
        if line.startswith("Good signature from "):
            return True
    return False


# The main parsing function to use in most cases:


def parse(input_stream=None, replace_errors=True):
    if input_stream is None:
        input_stream = sys.stdin
    if replace_errors:
        _reconfigure(input_stream)
    lines = input_stream_to_lines(input_stream)
    raw_commits = lines_to_raw_commits(lines)
    split_commits = raw_commits_to_split_commits(raw_commits)
    pretty_commits = split_commits_to_pretty_commits(split_commits)
    for pretty_commit in pretty_commits:
        # TODO: Do we want to do something more here?
        yield pretty_commit


# Another version of parse() which gives you all 3 representations of a commit
# Mostly useful for debugging
def parse_to_all_representations(input_stream=None, replace_errors=True):
    if input_stream is None:
        input_stream = sys.stdin
    if replace_errors:
        _reconfigure(input_stream)
    lines = input_stream_to_lines(input_stream)
    raw_commits = lines_to_raw_commits(lines)
    for raw_commit in raw_commits:
        # These function normally operate on iterators which yield many elements
        # (the entire git log), but we can call them on a n=1 tuple as well:
        split_commit = next(raw_commits_to_split_commits((raw_commit,)))
        pretty_commit = next(split_commits_to_pretty_commits((split_commit,)))
        yield (raw_commit, split_commit, pretty_commit)


# The individual steps / iterators:


def input_stream_to_lines(input_stream):
    for line in input_stream:
        yield line


def lines_to_raw_commits(line_iterator):
    current_commit = []
    for line in line_iterator:
        assert line[-1] == "\n"
        line = line[0:-1]
        if line.startswith("commit "):
            # New commit, finalize previous
            if current_commit:
                yield current_commit
                current_commit = []
        current_commit.append(line)
    if current_commit:
        yield current_commit


def raw_commits_to_split_commits(raw_commits):
    for raw in raw_commits:
        commit = {}
        first_line = raw[0]
        assert len(first_line) == len("commit 680e160eef58249b1b896512d50f6342ad325f01")
        assert first_line.startswith("commit ")
        commit["commit"] = [first_line]
        lines = (line for line in raw[1:])
        line = next(lines)
        while line and line.startswith("gpg: "):
            if "gpg" not in commit:
                commit["gpg"] = []
            commit["gpg"].append(line)
            line = next(lines)

        if line and line.startswith("Primary key fingerprint: "):
            commit["Primary key fingerprint"] = [line]
            line = next(lines)

        if line and line.startswith("     Subkey fingerprint: "):
            commit["Subkey fingerprint"] = [line]
            line = next(lines)

        assert line and line.startswith("tree ")  # TODO: Always?
        if line and line.startswith("tree "):
            commit["tree"] = [line]
            line = next(lines)

        while line and line.startswith("parent "):
            if "parent" not in commit:
                commit["parent"] = []
            commit["parent"].append(line)
            line = next(lines)

        assert line and line.startswith("author ")  # TODO: Always?
        if line and line.startswith("author "):
            commit["author"] = [line]
            line = next(lines)

        assert line and line.startswith("committer ")  # TODO: Always?
        if line and line.startswith("committer "):
            commit["committer"] = [line]
            line = next(lines)

        if line and line.startswith("gpgsig "):
            commit["gpgsig"] = [line]
            line = next(lines)
            while not line.startswith(" -----END PGP SIGNATURE-----"):
                assert "BEGIN PGP SIGNATURE" not in line
                commit["gpgsig"].append(line)
                line = next(lines)
            assert line == " -----END PGP SIGNATURE-----"
            commit["gpgsig"].append(line)
            line = next(lines)
        try:
            while line == " ":
                line = next(lines)
            while line == "":
                line = next(lines)
            commit["message"] = []
            while line and line != "---":
                commit["message"].append(line)
                line = next(lines)
            if line and line == "---":
                commit["diff"] = []
                line = next(lines)
                for line in lines:
                    commit["diff"].append(line)
        except StopIteration:
            pass
        yield commit


def _remove_prefixes(commit):
    if "commit" in commit:
        commit["commit"][0] = remove_prefix(commit["commit"][0], "commit ")
    if "gpg" in commit:
        commit["gpg"] = [remove_prefix(x, "gpg: ") for x in commit["gpg"]]
    if "Primary key fingerprint" in commit:
        commit["Primary key fingerprint"][0] = remove_prefix(
            commit["Primary key fingerprint"][0], "Primary key fingerprint: "
        )
    if "Subkey fingerprint" in commit:
        commit["Subkey fingerprint"][0] = remove_prefix(
            commit["Subkey fingerprint"][0], "     Subkey fingerprint: "
        )
    if "tree" in commit:
        commit["tree"][0] = remove_prefix(commit["tree"][0], "tree ")
    if "parent" in commit:
        commit["parent"][0] = remove_prefix(commit["parent"][0], "parent ")
    if "author" in commit:
        commit["author"][0] = remove_prefix(commit["author"][0], "author ")
    if "committer" in commit:
        commit["committer"][0] = remove_prefix(commit["committer"][0], "committer ")
    if "gpgsig" in commit:
        commit["gpgsig"][0] = remove_prefix(commit["gpgsig"][0], "gpgsig ")


def _strip_lists(commit):
    assert "commit" in commit and len(commit["commit"]) == 1
    commit["commit"] = commit["commit"][0]
    assert "tree" in commit and len(commit["tree"]) == 1
    commit["tree"] = commit["tree"][0]


def split_commits_to_pretty_commits(split_commits):
    for commit in split_commits:
        final = {}
        for key, value in commit.items():
            final[key] = value
        _remove_prefixes(final)
        _strip_lists(final)
        if "gpgsig" in final:
            assert final["gpgsig"][0] == "-----BEGIN PGP SIGNATURE-----"
            signature = ["-----BEGIN PGP SIGNATURE-----"]
            for line in final["gpgsig"][1:]:
                assert line.startswith(" ")
                signature.append(line[1:])
            final["gpgsig"] = "\n".join(signature)

        assert "author" in final
        final["author"] = parse_author(final["author"][0])
        assert "committer" in final
        final["committer"] = parse_author(final["committer"][0])

        if "Primary key fingerprint" in final:
            del final["Primary key fingerprint"]

        if "gpg" in final:
            if final["gpg"][0].startswith("Signature made "):
                final["valid_signature"] = True
                if not valid_signature(final):
                    final["valid_signature"] = False
                final["fingerprint"] = remove_prefix(
                    final["gpg"][1], "               using RSA key "
                )
                if "Primary key fingerprint" in final:
                    del final["Primary key fingerprint"]
            else:
                pass  # TODO look into this

        assert "message" in final
        final["message"] = "\n".join(x[4:] for x in final["message"])
        yield final



import sys

sys.path.append(".")

from glrp.pretty import pretty
import glrp.internal_parser as internal_parser


def test_basic():
    with open("tests/first_commits.txt", "r") as f:
        lines = f.readlines()
    content = "".join(lines)
    result = internal_parser.parse_to_all_representations(lines, replace_errors=False)
    result = list(result)
    assert len(result) == 2
    assert len(result[0]) == 3
    raw_content = ""
    for raw_commit, split_commit, pretty_commit in result:
        raw_content += "\n".join(raw_commit) + "\n"
    assert raw_content == content

    expected = []
    with open("tests/expected-01.json", "r") as f:
        expected.append(f.read())
    with open("tests/expected-02.json", "r") as f:
        expected.append(f.read())
    end_result = internal_parser.parse(lines, replace_errors=False)
    end_result = [pretty(x) + "\n" for x in end_result]

    assert len(end_result) == 2
    assert end_result[0] == expected[0]
    assert end_result[1] == expected[1]



import os
import json
from copy import copy, deepcopy
from typing import List


class Person:
    def __init__(self, id_string: str):
        assert type(id_string) is str
        assert len(id_string) > 0
        assert id_string.count(" <") == 1, "Invalid id_string"
        assert id_string.count(">") == 1, "Invalid id_string"
        self.id = id_string
        self.name, self.email = id_string.split(" <")
        self.email = self.email[:-1]
        assert len(self.name) > 0
        assert len(self.email) > 0


class Commit:
    def __init__(
        self,
        author: Person,
        committer: Person,
        message: str,
        fingerprint: str | None = None,
    ):
        self.author = author
        self.committer = committer
        self.message = message
        self.fingerprint = fingerprint

    @staticmethod
    def from_json(json_obj):
        author = Person(json_obj["author"]["id"])
        committer = Person(json_obj["committer"]["id"])
        message = json_obj["message"]
        fingerprint = json_obj.get("fingerprint", None)
        return Commit(author, committer, message, fingerprint)


def sum_int_dicts(a, b):
    r = {}
    for k in a:
        assert type(a[k]) is int
        r[k] = a[k]
        if k in b:
            assert type(b[k]) is int
            r[k] += b[k]
    # Keys unique to b:
    for k in (k for k in b if k not in a):
        assert type(b[k]) is int
        r[k] = b[k]
    return r


def add_entries(a, b):
    # "commits": 1,
    # "names": [commit.committer.name],
    # "ids": [commit.committer.id],
    # "fingerprints": [commit.fingerprint],
    a, b = deepcopy(a), deepcopy(b)
    r = {}
    common = []
    # Add everyting from a
    for k, v in a.items():
        r[k] = v
        if k in b:
            common.append(k)
    # Add unique items from b:
    for k, v in b.items():
        if k not in a:
            r[k] = v
        else:
            assert k in common
    # Combine / add comon items:
    for k in common:
        if k == "counts":
            r[k] = sum_int_dicts(a[k], b[k])
        else:
            # All others are lists of strings:
            for s in b[k]:
                if s not in r[k]:
                    r[k].append(s)
    return r


def counts(commits=0, signed=0, unsigned=0, trusted=0, untrusted=0):
    return {
        "commits": commits,
        "signed": signed,
        "unsigned": unsigned,
        "trusted": trusted,
        "untrusted": untrusted,
    }


class CommitSummary:
    def __init__(
        self,
        commit: Commit | None = None,
        filename: str | None = None,
        trusted: List[str] | None = None,
    ):
        """CommitSummary summarizes 0 or more commits.

        It can be created in a few different ways:

        1. Empty (A summary of 0 commits).
        2. Read in commit summary from JSON file (can be a summary of 0, 1, or
           more commits). Assumed to be the correct format, as generated by
           str(summary) or summary.to_dict().
        3. Based on a single commit object (summary of 1 commit).

        The typical way to use a CommitSummary to summarize multiple commits is to use
        the add (+) operator:

        summary = CommitSummary(a) + CommitSummary(b)

        This would create a summary of Commit a and b.
        """

        assert commit is None or filename is None
        assert commit is None or isinstance(commit, Commit)
        assert filename is None or isinstance(filename, str)
        assert filename is None or os.path.isfile(filename)

        self.counts = counts()  # 0 commits, 0 signed commits, etc.
        self.emails = {}
        self.names = {}
        self.ids = {}
        self.fingerprints = {}

        if filename:
            assert not commit
            with open(filename, "r") as f:
                data = json.load(f)
            self.counts = data["counts"]
            self.emails = data["emails"]
            self.names = data["names"]
            self.ids = data["ids"]
            self.fingerprints = data["fingerprints"]
            return

        if not commit:
            return

        assert not filename
        self.counts["commits"] += 1
        fp = commit.fingerprint
        if fp:
            self.counts["signed"] += 1
            if trusted and fp in trusted:
                self.counts["trusted"] += 1
            else:
                self.counts["untrusted"] += 1
        else:
            self.counts["unsigned"] += 1
        self.emails[commit.committer.email] = {
            "counts": copy(self.counts),
            "names": [commit.committer.name],
            "ids": [commit.committer.id],
            "fingerprints": [fp] if fp else [],
        }
        self.names[commit.committer.name] = {
            "counts": copy(self.counts),
            "emails": [commit.committer.email],
            "ids": [commit.committer.id],
            "fingerprints": [fp] if fp else [],
        }
        self.ids[commit.committer.id] = {
            "counts": copy(self.counts),
            "names": [commit.committer.name],
            "emails": [commit.committer.email],
            "fingerprints": [fp] if fp else [],
        }
        if fp:
            self.fingerprints[fp] = {
                "counts": copy(self.counts),
                "names": [commit.committer.name],
                "emails": [commit.committer.email],
                "ids": [commit.committer.id],
            }
        if commit.committer.id != commit.author.id:
            self.emails[commit.author.email] = {
                "counts": copy(self.counts),
                "names": [commit.author.name],
                "ids": [commit.author.id],
            }
            self.names[commit.author.name] = {
                "counts": copy(self.counts),
                "emails": [commit.author.email],
                "ids": [commit.author.id],
            }
            self.ids[commit.author.id] = {
                "counts": copy(self.counts),
                "names": [commit.author.name],
                "emails": [commit.author.email],
            }

    def add_update(self, other):
        assert isinstance(other, CommitSummary)
        for k, v in other.counts.items():
            if k not in self.counts:
                self.counts[k] = 0
            self.counts[k] += v
        for email, data in other.emails.items():
            if email not in self.emails:
                self.emails[email] = data
            else:
                self.emails[email] = add_entries(self.emails[email], data)
        for name, data in other.names.items():
            if name not in self.names:
                self.names[name] = data
            else:
                self.names[name] = add_entries(self.names[name], data)
        for id, data in other.ids.items():
            if id not in self.ids:
                self.ids[id] = data
            else:
                self.ids[id] = add_entries(self.ids[id], data)
        for fingerprint, data in other.fingerprints.items():
            if fingerprint not in self.fingerprints:
                self.fingerprints[fingerprint] = data
            else:
                self.fingerprints[fingerprint] = add_entries(
                    self.fingerprints[fingerprint], data
                )

        pass

    def __add__(self, other):
        result = CommitSummary()
        result.add_update(self)
        result.add_update(other)
        return result

    def to_dict(self):
        return {
            "counts": copy(self.counts),
            "emails": self.emails,
            "names": self.names,
            "fingerprints": self.fingerprints,
            "ids": self.ids,
        }

    def __str__(self):
        return json.dumps(self.to_dict())



def compare_summaries(before, after):
    for fingerprint in after["fingerprints"]:
        if fingerprint not in before["fingerprints"]:
            id = after["fingerprints"][fingerprint]["ids"][0]
            print(f"Note: New fingerprint {fingerprint} ({id})")
    for id in after["ids"]:
        if id not in before["ids"]:
            print(f"Note: New ID {id}")

    # Look through emails, warn about changes in fingerprints:
    for email in (e for e in after["emails"] if e in before["emails"]):
        if "fingerprints" not in after["emails"][email]:
            continue
        for fingerprint in after["emails"][email]["fingerprints"]:
            if (
                "fingerprints" not in before["emails"][email]
                or len(before["emails"][email]["fingerprints"]) == 0
            ):
                print(f"Note: First fingerprint for email {email}")
                continue
            if fingerprint not in before["emails"][email]["fingerprints"]:
                print(f"Warning: Changed fingerprint {fingerprint}")

    # Look through emails, warn about changes in names:
    for email in (e for e in after["emails"] if e in before["emails"]):
        for name in after["emails"][email]["names"]:
            if name not in before["emails"][email]["names"]:
                print(f"Warning: New name for {email} - {name}")

    # Look through fingerprints, warn about changes in names:
    for fingerprint in (
        e for e in after["fingerprints"] if e in before["fingerprints"]
    ):
        for id in after["fingerprints"][fingerprint]["ids"]:
            if id not in before["fingerprints"][fingerprint]["ids"]:
                print(f"Warning: New ID for {fingerprint} - {id}")



import os

files = os.listdir(__path__[0])
modules = (
    x.replace(".py", "") for x in files if x.endswith(".py") and not x.startswith("__")
)
for module in modules:
    __import__("glrp." + module)



import setuptools
import subprocess
import os

glrp_version = (
    subprocess.run(["git", "describe", "--tags"], stdout=subprocess.PIPE)
    .stdout.decode("utf-8")
    .strip()
)

if "-" in glrp_version:
    # when not on tag, git describe outputs: "1.3.3-22-gdf81228"
    # pip has gotten strict with version numbers
    # so change it to: "1.3.3+22.git.gdf81228"
    # See: https://peps.python.org/pep-0440/#local-version-segments
    v, i, s = glrp_version.split("-")
    glrp_version = v + "+" + i + ".git." + s

assert "-" not in glrp_version
assert "." in glrp_version

assert os.path.isfile("glrp/version.py")
with open("glrp/VERSION", "w", encoding="utf-8") as fh:
    fh.write("%s\n" % glrp_version)

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="glrp",
    version=glrp_version,
    author="Ole Herman Schumacher Elgesem",
    author_email="",
    description="Parser for git log --raw",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/olehermanse/glrp",
    packages=setuptools.find_packages(),
    package_data={"glrp": ["VERSION"]},
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.10",
    entry_points={"console_scripts": ["glrp = glrp.cli:main"]},
    install_requires=[],
)



# Taken from cfbs and simplified:
# https://github.com/cfengine/cfbs/blob/master/cfbs/pretty.py

# MIT License

# Copyright (c) 2021 Northern.tech

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import json
from collections import OrderedDict


def pretty_check_file(filename):
    with open(filename) as f:
        s = f.read()
    o = json.loads(s, object_pairs_hook=OrderedDict)
    return s == pretty(o) + "\n"


def pretty_check_string(s):
    o = json.loads(s, object_pairs_hook=OrderedDict)
    return s == pretty(o)


def pretty_file(filename):
    with open(filename) as f:
        old_data = f.read()
    new_data = pretty_string(old_data) + "\n"
    if old_data != new_data:
        with open(filename, "w") as f:
            f.write(new_data)
        return True
    return False


def pretty_string(s):
    s = json.loads(s, object_pairs_hook=OrderedDict)
    return pretty(s)


def pretty(o):
    MAX_LEN = 80
    INDENT_SIZE = 2

    def _should_wrap(parent, indent):
        assert isinstance(parent, (tuple, list, dict))
        # We should wrap the top level collection
        if indent == 0:
            return True
        if isinstance(parent, dict):
            parent = parent.values()

        count = 0
        for child in parent:
            if isinstance(child, (tuple, list, dict)):
                if len(child) >= 2:
                    count += 1
        return count >= 2

    def _encode_list(lst, indent, cursor):
        if not lst:
            return "[]"
        if not _should_wrap(lst, indent):
            buf = json.dumps(lst)
            assert "\n" not in buf
            if indent + cursor + len(buf) <= MAX_LEN:
                return buf

        indent += INDENT_SIZE
        buf = "[\n" + " " * indent
        first = True
        for value in lst:
            if first:
                first = False
            else:
                buf += ",\n" + " " * indent
            buf += _encode(value, indent, 0)
        indent -= INDENT_SIZE
        buf += "\n" + " " * indent + "]"

        return buf

    def _encode_dict(dct, indent, cursor):
        if not dct:
            return "{}"
        if not _should_wrap(dct, indent):
            buf = json.dumps(dct)
            buf = "{ " + buf[1 : len(buf) - 1] + " }"
            assert "\n" not in buf
            if indent + cursor + len(buf) <= MAX_LEN:
                return buf

        indent += INDENT_SIZE
        buf = "{\n" + " " * indent
        first = True
        for key, value in dct.items():
            if first:
                first = False
            else:
                buf += ",\n" + " " * indent
            if not isinstance(key, str):
                raise ValueError("Illegal key type '" + type(key).__name__ + "'")
            entry = '"' + key + '": '
            buf += entry + _encode(value, indent, len(entry))
        indent -= INDENT_SIZE
        buf += "\n" + " " * indent + "}"

        return buf

    def _encode(data, indent, cursor):
        if data is None:
            return "null"
        elif data is True:
            return "true"
        elif data is False:
            return "false"
        elif isinstance(data, (int, float)):
            return repr(data)
        elif isinstance(data, str):
            # Use the json module to escape the string with backslashes:
            return json.dumps(data)
        elif isinstance(data, (list, tuple)):
            return _encode_list(data, indent, cursor)
        elif isinstance(data, dict):
            return _encode_dict(data, indent, cursor)
        else:
            raise ValueError("Illegal value type '" + type(data).__name__ + "'")

    return _encode(o, 0, 0)
